from typing import Optional, Union

from fieldmarshal import Registry, Hook, struct, unmarshal
from pytest import raises as assert_raises


def test_unmarshal_union():
    assert unmarshal(1, Union[int, str]) == 1
    assert unmarshal('x', Union[int, str]) == 'x'


def test_unmarshal_optional():
    assert unmarshal(1, Optional[int]) == 1
    assert unmarshal(None, Optional[int]) is None


def test_unmarshal_union_no_hooks():
    @struct
    class Foo:
        value: int

    @struct
    class Bar:
        name: str

    assert unmarshal(1, Union[int, Foo]) == 1
    assert unmarshal(1, Union[int, Foo, Bar]) == 1
    assert unmarshal({'value': 1}, Union[int, Foo]) == Foo(1)

    with assert_raises(TypeError):
        unmarshal({'value': 1}, Union[Foo, Bar])

    assert unmarshal(None, Optional[Union[int, Foo]]) is None
    assert unmarshal(None, Optional[Union[Foo, Bar]]) is None


def test_resolve_union():
    @struct
    class Foo:
        value: int

    @struct
    class Bar:
        name: str

    def unmarshal_union_foo_bar(obj, type_hint, registry):
        if 'value' in obj:
            return registry.unmarshal(obj, Foo)
        return registry.unmarshal(obj, Bar)

    r = Registry()
    r.add_unmarshal_hook(Union[Foo, Bar], Hook(unmarshal_union_foo_bar))

    assert r.unmarshal({'value': 1}, Union[Foo, Bar]) == Foo(1)
    assert r.unmarshal({'name': 'x'}, Union[Foo, Bar]) == Bar('x')
    assert r.unmarshal({'value': 1}, Union[Bar, Foo]) == Foo(1)
    assert r.unmarshal({'name': 'x'}, Union[Bar, Foo]) == Bar('x')

    assert r.unmarshal(None, Optional[Union[Foo, Bar]]) is None
    assert r.unmarshal({'name': 'x'}, Union[Foo, Bar]) == Bar('x')
    assert r.unmarshal({'name': 'x'}, Optional[Union[Foo, Bar]]) == Bar('x')


def test_cant_unmarshal_union_types_having_hook():
    @struct
    class Foo:
        value: int

    class FooSubclass(Foo): pass

    r = Registry()
    r.add_unmarshal_hook(Foo, lambda v: Foo(v))

    assert r.unmarshal(1, Foo) == Foo(1)

    with assert_raises(TypeError):
        r.unmarshal(1, Union[int, Foo])

    with assert_raises(TypeError):
        r.unmarshal(1, Union[int, FooSubclass])

    assert r.unmarshal(None, Optional[Union[int, Foo]]) is None
    assert r.unmarshal(None, Optional[Union[int, FooSubclass]]) is None



from typing import List

import attr
import pytest
from fieldmarshal import Registry, struct, marshal


@pytest.mark.parametrize('value', [
    1, 0, 0.1, 0.0, True, False, None,
])
def test_marshal_simple_value(value):
    assert marshal(value) is value


@pytest.mark.parametrize('value', [
    [1, 2, 3],
    (1, 2, 3),
    set([1, 2, 3]),
    frozenset([1, 2, 3]),
])
def test_marshal_list_tuple_set(value):
    assert marshal(value) == [1, 2, 3]


@pytest.mark.parametrize('value, result', [
    ({'a': 1}, {'a': 1}),
    ({1: 'a'}, {'1': 'a'}),
    ({True: 1, False: 0}, {'true': 1, 'false': 0}),
    ({None: 0}, {'null': 0}),
])
def test_marshal_dict(value, result):
    assert marshal(value) == result


@pytest.mark.skip
@pytest.mark.parametrize('value, type_hint', [
    ([1], List[int]),
    (['a'], List[str]),
    ([0.0], List[float]),
    ([True], List[bool]),
    ([None], List[type(None)]),
])
def test_marshal_list_type_hint(value, type_hint):
    assert marshal(value, type_hint) is value


def test_marshal_struct():
    @struct
    class Foo:
        value: int

    assert marshal(Foo(1)) == {'value': 1}


def test_marshal_struct_nested():
    @struct
    class Foo:
        @struct
        class Bar:
            value: int
        bar: Bar

    assert marshal(Foo(Foo.Bar(2))) == {'bar': {'value':2}}


def test_marshal_class():
    class Foo:
        def __init__(self, value):
            self.value = value

    r = Registry()
    r.add_marshal_hook(Foo, lambda f: {'value': f.value})
    assert r.marshal(Foo(1)) == {'value': 1}



from typing import Any, List, Tuple, Set, FrozenSet, Dict, Optional

import attr
import pytest
from fieldmarshal import Registry, struct, unmarshal


@pytest.mark.parametrize('value', [
    1, 0, 0.1, 0.0, True, False, None,
])
def test_unmarshal_simple_value(value):
    assert unmarshal(value, type(value)) is value


@pytest.mark.parametrize('value', [
    1, 0, 0.1, 0.0, True, False, None,
])
def test_unmarshal_simple_value_any(value):
    assert unmarshal(value, Any) is value


@pytest.mark.parametrize('type_hint, result', [
    (List[int], [1, 2]),
    (Tuple[int, int], (1, 2)),
    (Set[int], {1, 2}),
    (FrozenSet[int], frozenset([1, 2])),
])
def test_unmarshal_list(type_hint, result):
    assert unmarshal([1, 2], type_hint) == result


@pytest.mark.parametrize('value, type_hint, result', [
    ({'a': 1}, Dict[str, int], {'a': 1}),
    ({'1': 'a'}, Dict[int, str], {1: 'a'}),
    ({'true': 1, 'false': 0}, Dict[bool, int], {True: 1, False: 0}),
    ({'null': 0}, Dict[type(None), int], {None: 0}),
])
def test_unmarshal_dict(value, type_hint, result):
    assert unmarshal(value, type_hint) == result


def test_unmarshal_struct():
    @struct
    class Foo:
        value: int

    assert unmarshal({'value': 1}, Foo) == Foo(1)


def test_unmarshal_struct_nested():
    @struct
    class Foo:
        @struct
        class Bar:
            value: int
        bar: Bar

    assert unmarshal({'bar': {'value': 2}}, Foo) == Foo(Foo.Bar(2))


def test_unmarshal_class():
    class Foo:
        def __init__(self, value):
            self.value = value

    r = Registry()
    r.add_unmarshal_hook(Foo, lambda d: Foo(value=d['value']))
    assert r.unmarshal({'value': 1}, Foo).value == 1
    assert r.unmarshal({'value': 1}, Optional[Foo]).value == 1




from fieldmarshal import Registry, struct, field, marshal, unmarshal
from pytest import raises as assert_raises


def test_rename_field():
    @struct
    class Foo:
        my_field: int = field('my-field')

    assert marshal(Foo(my_field=1)) == {'my-field': 1}
    assert unmarshal({'my-field': 1}, Foo) == Foo(my_field=1)


def test_omit_fields():
    @struct(auto_attribs=False)
    class Foo:
        a = field(omit_if_none=True)
        b = field(omit=True, default=None)

    assert marshal(Foo(a=1, b=2)) == {'a': 1}
    assert marshal(Foo(a=None, b=2)) == {}
    assert unmarshal({'a': 1, 'b': 2}, Foo) == Foo(a=1, b=None)


def test_missing_default():
    @struct
    class Foo:
        a: int
        b: int = 2

    assert unmarshal({'a': 1}, Foo) == Foo(1, 2)

    with assert_raises(KeyError):
        assert unmarshal({'b': 1}, Foo)


def test_field_hooks():
    @struct
    class Foo:
        a: str = field(
            marshal=lambda s: '<%s>' % s,
            unmarshal=lambda s: s.strip('<>'),
        )

    assert marshal(Foo('x')) == {'a': '<x>'}
    assert unmarshal({'a': '<x>'}, Foo) == Foo('x')


def test_field_hooks_override_registry_hooks():
    class MyString(str): pass

    @struct
    class Foo:
        a: MyString = field(
            marshal=lambda s: '<%s>' % s,
            unmarshal=lambda s: s.strip('<>'),
        )

    r = Registry()
    r.add_marshal_hook(MyString, lambda s: '[%s]' % s)
    r.add_unmarshal_hook(MyString, lambda s: s.strip('[]'))

    assert r.marshal(Foo(MyString('x'))) == {'a': '<x>'}
    assert r.unmarshal({'a': '<x>'}, Foo) == Foo('x')



from enum import Enum
from pytest import raises as assert_raises
from typing import Any, Dict, Union

from fieldmarshal import unmarshal


class MyEnum(Enum):
    A = 'a'


def test_union():
    # Union[int, str] does not really work as a dict key for JSON
    assert unmarshal({'a': 1}, Dict[Union[int, str], Any]) == {'a': 1}
    assert unmarshal({'1': 1}, Dict[Union[int, str], Any]) == {'1': 1}
    assert unmarshal({'a': 1}, Dict[Union[int, MyEnum], Any]) == {MyEnum.A: 1}


def test_bool_none():
    assert unmarshal({'true': 1}, Dict[bool, int]) == {True: 1}
    assert unmarshal({'null': 1}, Dict[type(None), int]) == {None: 1}

    with assert_raises(ValueError):
        unmarshal({'x': 1}, Dict[bool, int])

    with assert_raises(ValueError):
        unmarshal({'x': 1}, Dict[type(None), int])




import json
from enum import Enum, Flag, IntEnum, IntFlag
from functools import partial, singledispatch
from typing import Any, List, Tuple, Set, FrozenSet, Dict, Union

import attr

__version__ = '0.0.1'

__all__ = [
    'field',
    'marshal',
    'marshal_json',
    'struct',
    'unmarshal',
    'unmarshal_json'

    'Hook',
    'Options',
    'Registry',

    'DEFAULT_OPTIONS',
    'DEFAULT_REGISTRY',
]


struct = partial(attr.s, slots=True, auto_attribs=True)


@struct
class Options:
    """
    Field options control how a field is marshalled/unmarshalled.

    `name`
    :   Rename the field when marshalling/unmarshalling. Useful for example
    :   for JSON attribte names that are not valid Python identifiers.

    `omit`
    :   Ignore the field for the purpose of marshalling/unmarshalling. The
    :   field value will neither be read nor written to. The field should
    :   also have a default value for the class to support unmarshalling.

    `omit_if_none`
    :   Omit the field when marshalling if it's value is `None`. When
    :   unmarshalling, the field will be read as normal.

    `marshal`
    :   A function to call to marshal the contents of this field. The function
    :   will be passed the field value as its only argument. This hook
    :   overrides other marshal hooks registered for the field's type.

    `unmarshal`
    :   A function to call to unmarshal data for this field. The function
    :   will be passed the data bein unmarshalled as its only argument. This
    :   hook overrides other unmarshal hooks registered for the field's type.
    """
    name: str = None
    omit: bool = False
    omit_if_none: bool = False
    marshal: Any = None
    unmarshal: Any = None


def field(name=None, omit=False, omit_if_none=False, marshal=None, unmarshal=None, **kw):
    """
    Wrapper around `attr.ib` that accepts additional arguments.

    Arguments that control marshalling are saved as an `Options` object
    in the fields metadata under the 'fieldmarshal' key.
    """
    metadata = kw.setdefault('metadata', {})
    metadata['fieldmarshal'] = Options(
        name,
        omit=omit,
        omit_if_none=omit_if_none,
        marshal=marshal,
        unmarshal=unmarshal,
    )
    return attr.ib(**kw)


DEFAULT_OPTIONS  = Options()

IDENTITY = object()

NONE_TYPE = type(None)
SIMPLE_TYPES = (int, bool, float, str, NONE_TYPE)


def _marshal_default(obj, registry):
    raise TypeError("Can't marshal %r" % obj)


def _marshal_attrs(obj, registry):
    data = {}
    cls = obj.__class__
    for field in cls.__attrs_attrs__:
        key = (cls, field.name)
        try:
            options = registry._field_options_cache[key]
        except KeyError:
            options = field.metadata.get('fieldmarshal', DEFAULT_OPTIONS)
            registry._field_options_cache[key] = options
        if not options.omit:
            name = field.name if options.name is None else options.name
            value = getattr(obj, field.name)
            if not (value is None and options.omit_if_none):
                if options.marshal is not None:
                    data[name] = options.marshal(value)
                else:
                    data[name] = registry.marshal(value)
    return data


def _marshal_list(obj, registry):
    return [registry.marshal(item) for item in obj]


def _marshal_set(obj, registry):
    return sorted([registry.marshal(item) for item in obj])


def _marshal_dict_key(key, registry):
    obj = registry.marshal(key)

    # Rendering of True, False, None, object() as dict keys:
    #   stdlib json: "true", "false", "null", TypeError
    #   ujson:       "True", "False", "None", "<object object at …>"
    #   rapidjson:   TypeError (all)

    if obj is True:
        return 'true'
    elif obj is False:
        return 'false'
    elif obj is None:
        return 'null'
    elif isinstance(obj, str):
        return obj
    elif isinstance(obj, (int, float)):
        return str(obj)
    else:
        raise TypeError("Can't marshal dict key: %r" % key)


def _marshal_dict(obj, registry):
    return {_marshal_dict_key(k, registry): registry.marshal(v)
            for k, v in obj.items()}


def _marshal_enum(obj, registry):
    return obj.value


def _unmarshal_default(obj, type_hint, registry):
    raise TypeError("Can't unmarshal to %s: %r" % (type_hint, obj))


def _unmarshal_attrs(obj, type_hint, registry):
    kw = {}
    for field in type_hint.__attrs_attrs__:
        key = (type_hint, field.name)
        try:
            options = registry._field_options_cache[key]
        except KeyError:
            options = field.metadata.get('fieldmarshal', DEFAULT_OPTIONS)
            registry._field_options_cache[key] = options
        if not options.omit:
            name = field.name if options.name is None else options.name
            try:
                value = obj[name]
            except KeyError:
                if field.default is not attr.NOTHING:
                    value = field.default
                else:
                    raise
            if options.unmarshal is not None:
                kw[field.name] = options.unmarshal(value)
            else:
                kw[field.name] = registry.unmarshal(value, field.type or Any)
    return type_hint(**kw)


def _unmarshal_list(obj, type_hint, registry):
    # in Python 3.6, List[int].__origin__ is List
    if type_hint.__origin__ in (list, List):
        item_type, = type_hint.__args__
        if item_type in SIMPLE_TYPES:
            return obj
        else:
            return [registry.unmarshal(item, item_type) for item in obj]
    elif type_hint.__origin__ in (tuple, Tuple):
        item_types = type_hint.__args__
        if all([t in SIMPLE_TYPES for t in item_types]):
            return tuple(obj)
        else:
            return tuple([registry.unmarshal(item, type_)
                for item, type_ in zip(obj, item_types)])
    elif type_hint.__origin__ in (set, Set):
        item_type, = type_hint.__args__
        if item_type in SIMPLE_TYPES:
            return set(obj)
        else:
            return set([registry.unmarshal(item, item_type) for item in obj])
    elif type_hint.__origin__ in (frozenset, FrozenSet):
        item_type, = type_hint.__args__
        if item_type in SIMPLE_TYPES:
            return set(obj)
        else:
            return frozenset([registry.unmarshal(item, item_type) for item in obj])

    raise TypeError("Can't unmarshal to %s: %r" % (type_hint, obj))


def _unmarshal_dict_key(key, type_, registry):
    if type_ in (int, float):
        obj = type_(key)
    elif type_ is bool:
        try:
            obj = {'true': True, 'false': False}[key]
        except KeyError:
            raise ValueError(
                "Error converting dict key to bool. Expected 'true' or 'false', got %r" % key
            ) from None
    elif type_ is NONE_TYPE:
        try:
            obj = {'null': None}[key]
        except KeyError:
            raise ValueError(
                "Error converting dict key to NoneType. Expected 'null', got %r" % key
            ) from None
    elif getattr(type_, '__mro__', None) is not None and issubclass(type_, (Flag, IntEnum, IntFlag)):
        obj = int(key)
    else:
        obj = key

    return registry.unmarshal(obj, type_)


def _unmarshal_dict(obj, type_hint, registry):
    if type_hint.__origin__ in (dict, Dict):
        key_type, value_type = type_hint.__args__
        if value_type in SIMPLE_TYPES:
            return {_unmarshal_dict_key(k, key_type, registry): v
                    for k, v in obj.items()}
        else:
            return {_unmarshal_dict_key(k, key_type, registry):
                        registry.unmarshal(v, value_type)
                    for k, v in obj.items()}

    raise TypeError("Can't unmarshal to %s: %r" % (type_hint, obj))


def _unmarshal_enum(obj, type_hint, registry):
    return type_hint(obj)


def _unmarshal_union(obj, type_hint, registry):
    cls = obj.__class__
    union_types = type_hint.__args__
    # Special-case Optional[]
    if NONE_TYPE in union_types:
        if obj is None:
            return None
        else:
            not_none = tuple(t for t in union_types if t is not NONE_TYPE)
            return registry.unmarshal(obj, Union[not_none])
    # TODO this is expensive and should be cached.
    if cls in union_types and not any([
            t for t in union_types
            if registry.lookup_unmarshal_impl(cls, t) in registry._unmarshal_hooks
        ]):
        return obj
    candidates = [t for t in union_types if t not in SIMPLE_TYPES]
    if len(candidates) == 1:
        type_ = candidates[0]
        if registry.lookup_unmarshal_impl(cls, type_) not in registry._unmarshal_hooks:
            return registry.unmarshal(obj, type_)

    raise TypeError("Can't unmarshal to %s: %r" % (type_hint, obj))


@struct
class Hook:
    """
    Wrapper around a marshal or unmarshal hook.

    Can be used to specify that the hook accepts additional arguments other
    than the object being marshalled/unmarshalled.

    When `takes_args` is True (the default), additional arguments will be
    passed to the hook. The type of arguments depends on the type of hook. See
    `Registry.add_marshal_hook` and `Registry.add_unmarshal_hook` for details.
    """
    fn: Any
    takes_args: bool = True


class Registry:
    def __init__(self):
        """
        Create a registry instance.

        A registry is used for marshalling and unmarshalling objects, and
        for registering hooks for types that are not handled natively.
        """
        self._marshal_impl_cache = {}
        self._unmarshal_impl_cache = {}
        self._marshal_impl_dispatch = singledispatch(_marshal_default)
        self._unmarshal_impl_dispatch = singledispatch(_unmarshal_default)
        self._unmarshal_hook_impl = {}
        self._unmarshal_hooks = set()
        self._field_options_cache = {}

        for type_ in SIMPLE_TYPES:
            self._marshal_impl_dispatch.register(type_, IDENTITY)

        self._marshal_impl_dispatch.register(list, _marshal_list)
        self._marshal_impl_dispatch.register(tuple, _marshal_list)
        self._marshal_impl_dispatch.register(set, _marshal_set)
        self._marshal_impl_dispatch.register(frozenset, _marshal_set)
        self._marshal_impl_dispatch.register(dict, _marshal_dict)
        self._marshal_impl_dispatch.register(Enum, _marshal_enum)
        self._marshal_impl_dispatch.register(IntEnum, _marshal_enum)
        self._marshal_impl_dispatch.register(IntFlag, _marshal_enum)

        for type_ in SIMPLE_TYPES:
            self._unmarshal_impl_dispatch.register(type_, IDENTITY)

        self._unmarshal_impl_dispatch.register(list, _unmarshal_list)
        self._unmarshal_impl_dispatch.register(tuple, _unmarshal_list)
        self._unmarshal_impl_dispatch.register(set, _unmarshal_list)
        self._unmarshal_impl_dispatch.register(frozenset, _unmarshal_list)
        self._unmarshal_impl_dispatch.register(dict, _unmarshal_dict)
        self._unmarshal_impl_dispatch.register(Enum, _unmarshal_enum)
        self._unmarshal_impl_dispatch.register(IntEnum, _unmarshal_enum)
        self._unmarshal_impl_dispatch.register(IntFlag, _unmarshal_enum)


    def marshal(self, obj):
        """
        Marshal an object to a JSON-compatible data structure.

        The resulting data structure contains only objects of type
        `list`, `dict` (with string keys), `int`, `float`, `str`, `bool` or
        `NoneType` and can be converted to JSON without further modifications.

        The reverse operation is `unmarshal`.
        """
        key = obj.__class__
        try:
            impl = self._marshal_impl_cache[key]
        except KeyError:
            impl = self.lookup_marshal_impl(key)
            self._marshal_impl_cache[key] = impl
        if impl is IDENTITY:
            return obj
        else:
            return impl(obj, self)

    def marshal_json(self, obj):
        """
        Marshal an object to a JSON string.

        Like `marshal`, but converts the result to JSON.
        """
        return json.dumps(self.marshal(obj))

    def add_marshal_hook(self, type_, fn):
        """
        Add a custom marshalling implementation for a type.

        `type_` can be a class or a concrete type from the `typing` module,
        such as `Union[list, str]`. The hook can either be a function that
        takes one argument (the object being marshalled), or a `Hook` object,
        which can be used to opt-in to receive the registry instance as an
        additional argument.

        The hook should return a JSON-compatible object. The hook will be
        called when an object of the specified type is encountered when
        marshalling. If `type_` is a class, the hook will also be used
        for instances of subclasses of `type_`, unless a more specific
        hook can be found.
        """
        hook = fn if isinstance(fn, Hook) else Hook(fn, False)
        if hook.takes_args:
            hook_impl = hook.fn
        else:
            hook_impl = lambda obj, _: hook.fn(obj)
        self._marshal_impl_dispatch.register(type_, hook_impl)
        self._marshal_impl_cache.clear()

    def lookup_marshal_impl(self, cls):
        """
        Return the marshal implementation that would be used for `cls`.
        """
        impl = self._marshal_impl_dispatch.dispatch(cls)
        if impl is _marshal_default and attr.has(cls):
            return _marshal_attrs
        return impl

    def unmarshal(self, obj, type_hint):
        """
        Unmarshal an object from a JSON-compatible data structure.

        The data structure must contain only objects of type `list`, `dict`
        (with string keys), `int`, `float`, `str`, `bool` or `NoneType`, such
        as returned by `marshal`.

        `type_hint` specifies the type of object to create. This can be a class
        or a concrete type from the `typing` module, such as `List[int]`.

        The reverse operation is `marshal`.
        """
        key = (obj.__class__, type_hint)
        try:
            impl = self._unmarshal_impl_cache[key]
        except KeyError:
            impl = self.lookup_unmarshal_impl(*key)
            self._unmarshal_impl_cache[key] = impl
        if impl is IDENTITY:
            return obj
        else:
            return impl(obj, type_hint, self)

    def unmarshal_json(self, data, type_hint):
        """
        Unmarshal an object from a JSON string.

        Like `unmarshal`, but accepts a JSON-encoded string.
        """
        return self.unmarshal(json.loads(data), type_hint)

    def add_unmarshal_hook(self, type_, fn):
        """
        Add a custom unmarshalling implementation for a type.

        `type_` can be a class or a concrete type from the `typing` module,
        such as `Union[list, str]`. The hook can either be a function that
        takes one argument (the object being unmarshalled), or a `Hook` object,
        which can be used to opt-in to receive the type the object is being
        unmarshalled to and the registry instance as additional arguments. The
        type passed this way is not necessarily the type the hook was
        registered for (it could be a subclass, for example).

        The hook will be called when data needs to be marshalled to an object
        of the specified type. If `type_` is a class, the hook will also be
        used for instances of subclasses of `type_`, unless a more specific
        hook can be found.
        """
        hook = fn if isinstance(fn, Hook) else Hook(fn, False)
        if hook.takes_args:
            hook_impl = hook.fn
        else:
            hook_impl = lambda obj, _, __: hook.fn(obj)
        self._unmarshal_hook_impl[type_] = hook_impl
        self._unmarshal_hooks.add(hook_impl)
        if getattr(type_, '__mro__', None) is not None:
            self._unmarshal_impl_dispatch.register(type_, hook)
            self._unmarshal_impl_cache.clear()

    def lookup_unmarshal_impl(self, cls, type_hint):
        """
        Return the unmarshal implementation that would be used for
        unmarshalling data of type `cls` to an object of type `type_hint`.
        """
        if type_hint is Any:
            return IDENTITY
        # TODO find impl for Type here if type_hint is Optional[Type]?
        if type_hint in self._unmarshal_hook_impl:
            return self._unmarshal_hook_impl[type_hint]
        elif getattr(type_hint, '__mro__', None) is not None:
            impl = self._unmarshal_impl_dispatch.dispatch(type_hint)
            if impl is _unmarshal_default and attr.has(type_hint):
                return _unmarshal_attrs
            return impl
        else:
            origin = getattr(type_hint, '__origin__', None)
            if origin is not None:
                if origin is Union:
                    return _unmarshal_union
                if getattr(origin, '__mro__', None) is not None:
                    return self._unmarshal_impl_dispatch.dispatch(origin)

        return _unmarshal_default

    def clear_cache(self):
        """
        Clear all caches of the registry.

        This should not be necessary unless classes are modified at runtime.
        """
        self._marshal_impl_cache.clear()
        self._unmarshal_impl_cache.clear()
        self._marshal_impl_dispatch._clear_cache()
        self._unmarshal_impl_dispatch._clear_cache()
        self._field_options_cache.clear()


DEFAULT_REGISTRY = Registry()

marshal = DEFAULT_REGISTRY.marshal
marshal_json = DEFAULT_REGISTRY.marshal_json
unmarshal = DEFAULT_REGISTRY.unmarshal
unmarshal_json = DEFAULT_REGISTRY.unmarshal_json



from enum import Enum, IntEnum, Flag, IntFlag, auto
from typing import Dict

from fieldmarshal import Registry, marshal, unmarshal


class MyStrEnum(Enum):
    A = 'a'


class MyBoolEnum(Enum):
    A = True


class MyFloatEnum(Enum):
    A = 0.1


class MyIntEnum(IntEnum):
    A = 1


class MyFlag(Flag):
    A = 1
    B = 2
    AB = 3


class MyIntFlag(IntFlag):
    A = auto()
    B = auto()


def test_marshal_enum():
    assert marshal(MyStrEnum.A) == 'a'
    assert marshal(MyBoolEnum.A) is True
    assert marshal(MyFloatEnum.A) == 0.1
    assert marshal(MyIntEnum.A) == 1
    assert marshal(MyFlag.A) == 1
    assert marshal(MyIntFlag.A) == 1


def test_marshal_flag():
    assert marshal(MyFlag.A & MyFlag.B) == 0
    assert marshal(MyFlag.A | MyFlag.B) == 3
    assert unmarshal(0, MyFlag) == MyFlag(0)
    assert unmarshal(3, MyFlag) == MyFlag.AB


def test_unmarshal_enum():
    assert unmarshal('a', MyStrEnum) is MyStrEnum.A
    assert unmarshal(True, MyBoolEnum) is MyBoolEnum.A
    assert unmarshal(0.1, MyFloatEnum) is MyFloatEnum.A
    assert unmarshal(1, MyIntEnum) is MyIntEnum.A
    assert unmarshal(1, MyFlag) is MyFlag.A
    assert unmarshal(1, MyIntFlag) is MyIntFlag.A


def test_marshal_enum_dict_keys():
    assert marshal({MyStrEnum.A: 1}) == {'a': 1}
    assert marshal({MyBoolEnum.A: 1}) == {'true': 1}
    assert marshal({MyFloatEnum.A: 1}) == {'0.1': 1}
    assert marshal({MyIntEnum.A: 1}) == {'1': 1}
    assert marshal({MyFlag.A: 1}) == {'1': 1}
    assert marshal({MyIntFlag.A: 1}) == {'1': 1}


def test_unmarshal_enum_dict_keys():
    assert unmarshal({'a': 1}, Dict[MyStrEnum, int]) == {MyStrEnum.A: 1}
    assert unmarshal({'1': 1}, Dict[MyIntEnum, int]) == {MyIntEnum.A: 1}
    assert unmarshal({'1': 1}, Dict[MyFlag, int]) == {MyFlag.A: 1}
    assert unmarshal({'1': 1}, Dict[MyIntFlag, int]) == {MyIntFlag.A: 1}


def test_unmarshal_enum_hooks():

    def unmarshal_bool_enum(v):
        return MyBoolEnum({'true': True}.get(v, v))

    def unmarshal_float_enum(v):
        return MyFloatEnum(float(v))

    r = Registry()
    r.add_unmarshal_hook(MyBoolEnum, unmarshal_bool_enum)
    r.add_unmarshal_hook(MyFloatEnum, unmarshal_float_enum)

    assert r.unmarshal('true', MyBoolEnum) is MyBoolEnum.A
    assert r.unmarshal({'true': 1}, Dict[MyBoolEnum, int]) == {MyBoolEnum.A: 1}

    assert r.unmarshal('0.1', MyFloatEnum) is MyFloatEnum.A
    assert r.unmarshal({'0.1': 1}, Dict[MyFloatEnum, int]) == {MyFloatEnum.A: 1}



from datetime import datetime
from enum import Enum, unique
from typing import Any, Dict, List, Union

from fieldmarshal import Registry, Hook, struct, field


@unique
class ServerStatus(Enum):
    RUNNING = 'running'
    STARTING = 'starting'
    STOPPING = 'stopping'
    OFF = 'off'
    UNKNOWN = 'unknown'


@struct
class ServerType:
    id: int
    name: str
    num_cores: int = field('cores')
    memory_gb: int = field('memory')
    disk_gb: int = field('disk')


@struct
class DnsPtr:
    ip: str
    dns_ptr: str


@struct
class IPAddress:
    ip: str
    dns_ptr: Union[str, List[DnsPtr]]


@struct
class Server:

    @struct
    class PublicNet:
        ipv4: IPAddress
        ipv6: IPAddress
        floating_ips: List[int]

    id: int
    name: str
    status: ServerStatus
    created: datetime
    server_type: ServerType
    public_net: PublicNet
    labels: Dict[str, str]


@struct
class ServerResponse:
    server: Server


def unmarshal_dns_ptr(obj, type_hint, registry):
    if isinstance(obj, str):
        return obj
    return registry.unmarshal(obj, List[DnsPtr])


registry = Registry()

# FIXME datetime.fromisoformat requires Python >= 3.7
registry.add_unmarshal_hook(datetime, datetime.fromisoformat)
registry.add_marshal_hook(datetime, lambda dt: dt.isoformat())

registry.add_unmarshal_hook(Union[str, List[DnsPtr]], Hook(unmarshal_dns_ptr))


def main(data):
    server = registry.unmarshal_json(data, ServerResponse).server
    #print(server)
    print('server.id: %r' % server.id)
    print('server.name: %r' % server.name)
    print('server.status: %r' % server.status)
    print('server.created: %r' % server.created)
    print('server.labels: %r' % server.labels)
    print('server.public_net.ipv4.ip: %r' % server.public_net.ipv4.ip)
    print('server.public_net.ipv4.dns_ptr: %r' % server.public_net.ipv4.dns_ptr)
    print('server.public_net.ipv6.ip: %r' % server.public_net.ipv6.ip)
    print('server.public_net.ipv6.dns_ptr[0].ip: %r' % server.public_net.ipv6.dns_ptr[0].ip)
    print('server.public_net.ipv6.dns_ptr[0].dns_ptr: %r' % server.public_net.ipv6.dns_ptr[0].dns_ptr)
    print('server.server_type.id: %r' % server.server_type.id)
    print('server.server_type.name: %r' % server.server_type.name)
    print('server.server_type.num_cores: %r' % server.server_type.num_cores)
    print('server.server_type.memory_gb: %r' % server.server_type.memory_gb)
    print('server.server_type.disk_gb: %r' % server.server_type.disk_gb)

    assert registry.unmarshal(registry.marshal(server), Server) == server


if __name__ == '__main__':
    import os
    file_path = os.path.join(os.path.dirname(__file__), 'server.json')
    with open(file_path) as f:
        main(f.read())



import os
import re

from setuptools import setup, find_packages

with open(os.path.join('src', 'fieldmarshal', '__init__.py')) as f:
    VERSION = re.search(
        r'^__version__\s*=\s*[\'"](.*)[\'"]', f.read(), re.M
    ).group(1)

with open('README.md') as f:
    README = f.read()

setup(
    name='fieldmarshal',
    version=VERSION,
    author='Stanis Trendelenburg',
    author_email='stanis.trendelenburg@gmail.com',
    url='https://github.com/trendels/fieldmarshal',
    license='MIT',
    description='Marshal/unmarshal attrs-based data classes to and from JSON',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers = [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=['attrs>=17.4.0'],
    zip_safe=False,
)



import attr
from fieldmarshal import Options, struct, field


def test_struct():
    @struct
    class Foo:
        value: int

    assert attr.has(Foo)
    assert attr.fields(Foo).value.type == int
    assert 'value' in Foo.__slots__


def test_struct_override():
    @struct(slots=False, auto_attribs=False)
    class Foo:
        value = attr.ib()

    assert attr.has(Foo)
    assert attr.fields(Foo).value.type is None
    assert getattr(Foo, '__slots__', None) is None


def test_field():
    @struct
    class Foo:
        value: int = field('v')

    assert attr.fields(Foo).value.metadata['fieldmarshal'] == Options(name='v')
